/*--------------------------------------------------------------------
 * $Id: mie_lex.l,v 1.4 2001/06/20 14:30:58 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

%{

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include "./mie.h"


int mie_check (input_struct input);

float  yytext2float(char* yytext, int which_token, int max_token);
float* yytext2floats(char* yytext, int nary);
int    yytext2nfloats(char* yytext);
int    yytext2int(char* yytext, int which_token, int max_token);
int*   yytext2integers(char* yytext, int nary);
int    yytext2nintegers(char* yytext);
char*  yytext2string(char* yytext, int which_token, int max_token);

char* s, *t;
float* tmp;
int nary;
int ierror=0;
int start_condition = INITIAL;
int line_number;
input_struct Input;
output_struct Output;

%}

%option noyywrap

DIGIT    [0-9]
SPACE    [ \t]
CCHAR    [#]
COMMENT  ({CCHAR}.*)
INTEGER  {DIGIT}+
SPACE_INTEGER {SPACE}+{INTEGER}
INTEGERS {SPACE_INTEGER}+ 
EXPON   ([DdEe][+-]?{DIGIT}+)
FLOAT    ((({DIGIT}+"."{DIGIT}*)|({DIGIT}+"."{DIGIT}*{EXPON}))|({INTEGER}))
SPACE_FLOAT {SPACE}+{FLOAT}
FLOATS   {SPACE_FLOAT}+ 
SIGN     [-+]
SIGNED_FLOAT {SIGN}?{FLOAT}
SPACE_SIGNED_FLOAT {SPACE}+{SIGNED_FLOAT}
SIGNED_FLOATS {SPACE_SIGNED_FLOAT}+
SIGNED_INTEGER {SIGN}?{INTEGER}
WORD     [_\.\/A-Za-z0-9][_\.\/A-Za-z0-9\-]*

ZERO_OR_TWO_SIGNED_FLOATS ({SPACE}?{SIGNED_FLOAT}+{SPACE}+{SIGNED_FLOAT})?
ZERO_OR_TWO_OR_WORD {ZERO_OR_TWO_SIGNED_FLOATS}|{WORD}
%%

  /* These are the allowed input lines in the mie input file 
     The input line specifications must come before the eating of
     blanks, otherwise things get messed up 
     Obviously this whole lot could have been handled more properly
     using bison etc., but than rather few would know what was going 
     on??? 
   */

mie_program{SPACE}+{WORD}           {s =  yytext2string(yytext,2,2);
				     if (strncmp("MIEV0",s,5)==0) {Input.program = MIEV0;}
				     else if (strncmp(s,"BH",2)==0) {Input.program = BHMIE;}}
mimcut{SPACE}+{SIGNED_FLOAT}        {Input.mie.mimcut   = yytext2float(yytext,2,2); }
nmom{SPACE}+{INTEGER}               {Input.mie.nmom  = yytext2int(yytext,2,2); 
                                     Input.mie.momdim = Input.mie.nmom+1; }
r_mean{SPACE}+{SIGNED_FLOAT}        {Input.r_mean   = yytext2float(yytext,2,2); }

refrac{SPACE}+{WORD}{SPACE}?{ZERO_OR_TWO_OR_WORD}  {
                                       s = (char *) calloc (strlen(yytext)+1,sizeof(char));
                                       strcpy (s, yytext); 
				       t =  yytext2string(s,2,4);
				       if (strncmp("ice",t,3)==0) 
					 Input.medium = ICE;
				       else  {
					 if (strncmp("water",t,5)==0) 
					   Input.medium = WATER;
					 else  {
					   if (strncmp("user",t,4)==0) {
					     Input.medium = USER;
					     
					     strcpy (s, yytext); 
					     nary = yytext2nfloats(s);
					     
					     strcpy (s, yytext);
					     tmp = yytext2floats(s, nary);
					     
					     Input.crefin.re = tmp[1];
					     Input.crefin.im = tmp[2];
					   }
					   else {
					     if (strncmp("file",t,4)==0) {
					       Input.medium = USER;
					       Input.ri_filename = yytext2string(yytext,3,3); 
					     }
					   }
					 }
				       }
				     
				     free(s);}
size_distribution_file{SPACE}+{WORD} {Input.sd_filename = yytext2string(yytext,2,2); }
wvn{FLOATS}                         {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.wl.wl_start  = yytext2float(s,2,3); 
                                     Input.wl.wl_end    = yytext2float(yytext,3,3);
				     free(s);}
wvn_step{SPACE}+{SIGNED_FLOAT}      {Input.wl.wl_step   = yytext2float(yytext,2,2); }

{SPACE}+       ;         /* eat up blanks */ 

{COMMENT}*     ;         /* eat up comments */

\n                        { ++line_number;};

{WORD}                    {fprintf(stderr,"Unknown command on line %d\n", line_number+1);
                           fprintf(stderr,"%s\n", yytext);
                           ierror++;}

<<EOF>> { 
  if ( ierror > 0 ) { exit(1); }
	    yyterminate();
        }      

%%

/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void usage()
{
  fprintf (stderr, "\nCalculates scattering and absorption properties using Mie theory.\n");
  fprintf (stderr, "See libRadtran documentation for more information.\n\n");
  fprintf (stderr, "No error checking yet, so be careful.\n\n");

  fprintf (stderr, "Usage: mie [-h] < input_file > Output_file\n");
  fprintf (stderr, " -h    Print this message.\n");  
  fprintf (stderr, "\n");  
}

int main(int argc, char **argv) 
{
  
  /*  @40c@                

@code{mie} performes Mie scattering calculations for a specified
wavelength interval. It reads input from standard input, 
and outputs to standard output. It is normally invoked in the following
way: 

@example
mie < input_file > output_file
@end example

@noindent The format of the input and output files are described
below. Several realistic examples of input files are subsequently 
given.

@strong{Warning:} Please note the error checking on input variables 
is very scarce at the moment. Hence, if you provide erroneous input, 
the outcome is unpredictable.
  @c40@ */

  /*  @40_01c@                
@cindex input file, mie
@cindex mie, input file

The mie input file consists of single line entries, each making
up a complete input to the mie program. First on the line comes
the parameter name, followed by one or more parameter values. The 
parameter name and the parameter values are seperated by white space.

Filenames are entered without any surrounding single or double quotes.

Comments are introduced by a @code{#}. Blank lines are ignored.

The various input parameters are described in detail below.
@vtable @code
@item mie_program
Specify which Mie program to use:
@vtable @code
@item BH
The Mie scattering program by Bohren and Hoffmann @*
(ftp://astro.princeton.edu/draine/scat/bhmie/)
@item MIEV0
The Mie scattering program by W. Wiscombe @*
(ftp://climate.gsfc.nasa.gov/pub/wiscombe).
@end vtable
@item mimcut
(positive) value below which imaginary refractive
index is regarded as zero (computation proceeds
faster for zero imaginary index). Only used by @code{mie_program} @code{MIEV0}.
@item nmom 
Number of moments of the phase function to be calculated (default: 0). 
Only possible with @code{mie_program} @code{MIEV0}.
@item r_mean
The radius [micron] of the particle to calculate single scattering properties of.
Used together with the wavelength information to calculate the 
Mie size parameter.
@item refrac
Specify which refractive index to use. The following options are implemented: 
@vtable @code
@item ice
The complex refractive index is taken from the REFICE function of W. Wiscombe. 
@item water
The complex refractive index is taken from the REFWAT function of W. Wiscombe. 
@item user <re> <im>
A user defined refractive index. re and im are the real and imaginary parts
(both positive numbers).
@item file <filename>
Read refractive index from a three-column file containing wavelength [nm],
and the real and imaginary parts of the refractive index (both positive numbers).
The Mie calculation is done for each wavelength defined here.
@end vtable
@item size_distribution_file
Specify a two column file, r [micron], n(r), which describes a size distribution 
of droplets. The Mie calculation is repeated for each value of r found in the 
size distribution file, and the final result is a weighted average of these 
values. The user himself therefore has to choose a set of r's suited for 
his specific purpose.
@item wvn
Sets the wavelength range, in nm. Specify first wavelength and last 
wavelength. The wavelength step is specified by @code{wvn_step}.
Ignored if @code{refrac file} is specified.
@item wvn_step
The wavelength step, in nm. Ignored if @code{refrac file} is specified.

@end vtable
  @c40_01@ */

  /*  @40_02c@                
@cindex output, mie
@cindex uvspec, mie
The mie output is currently fixed. We hope to make it user controllable
in the future.

mie outputs one line to standard output (stdout) for each wavelength.
The format of the output line is

@example
lambda refrac_real refrac_img qext omega gg spike pmom(0:nmom)
@end example

Here

@vtable @code
@item lambda
Wavelength, in nm.
@item refrac_real
The real part of the refractive index.
Meaningless if a @code{size_distribution_file} was specified.
@itemx refrac_img
The imaginary part of the refractive index.
Meaningless if a @code{size_distribution_file} was specified.
@item qext
The extinction efficiency factor if @code{r_mean} was specified or
the extinction coefficient [km-1] per unit liquid water content [g/m3]
if a @code{size_distribution_file} was specified.
@item omega
The single scattering albedo.
@item gg
The asymmetry parameter.
@item spike
To quote from Wiscombe's @file{MIEV0.doc}:
     @quotation
(REAL) magnitude of the smallest denominator of
            either Mie coefficient (a-sub-n or b-sub-n),
            taken over all terms in the Mie series past
            N = size parameter XX.  Values of SPIKE below
            about 0.3 signify a ripple spike, since these
            spikes are produced by abnormally small denominators
            in the Mie coefficients (normal denominators are of
            order unity or higher).  Defaults to 1.0 when not
            on a spike.  Does not identify all resonances
            (we are still working on that).
     @end quotation
Meaningless if a @code{size_distribution_file} was specified.
@item pmom(0:nmom)
The moments of the phase function. The phase function 
@iftex
 @tex
 $p(\mu)$ is @*
  $$ p (\mu) = \sum_{m=0}^{\infty} (2m+1) \cdot k_m \cdot P_m (\mu) $$
  where $k_m$ is the m'th moment and $P_m (\mu)$ is the m'th Legendre polynomial.
 @end tex
@end iftex
@ifinfo
 p(mu) is @*
 p (mu) = sum (m=0 to infinity) (2m+1) * k(m) * Pm (mu) @*
 where k(m) is the m'th moment and Pm (mu) is the m'th Legendre polynomial.
@end ifinfo
@end vtable

  @c40_02@ */

  /*  @40_03c@                

An example of a complete input file is

@example
@include ../examples/MIE_1.INP
@end example


      @c40_03@   */

  /***************** End of documentation *********************/


  /* prototype of main function */
  int mie(input_struct input, output_struct *output);

  int c;
  /* int iu, iv, j, lev; */
  int status=0;
  /*  Input_struct Input;
  Output_struct Output;*/

  /* Initialization */

  Output.print.phi    = 1; /* Default print radiances */


  Input.medium          = WATER;
  Input.r_mean          = 0;
  Input.sd_filename     = "";
  Input.ri_filename     = "";
  Input.program         = MIEV0;

  Input.wl.wl_start     = 0.0;
  Input.wl.wl_end       = 0.0;
  Input.mie.anyang      = 0;
  Input.mie.ipolzn      = 0;
  Input.mie.momdim      = 1;
  Input.mie.nmom        = 0;
  Input.mie.numang      = 0;
  Input.mie.mimcut      = 1.0e-08;
  Input.mie.perfct      = 0;
  Input.mie.prnt[0]     = 0;
  Input.mie.prnt[1]     = 0;
  Input.mie.xmu         = (float *) calloc (Input.mie.numang+1, sizeof(float));
  Input.mie.s1          = (mie_complex *) calloc (Input.mie.numang+1, sizeof(mie_complex));
  Input.mie.s2          = (mie_complex *) calloc (Input.mie.numang+1, sizeof(mie_complex));

  while ((c=getopt (argc, argv, "h")) != EOF)  {
    switch(c)  {
    case 'h': 
      usage();
      return (-1);
      break;
    default:
      usage();
      return (-1);
    }
  }

  /* Read the input file */

  yylex();

  status = mie_check(Input);
  if (status!=0) {
    fprintf (stderr, "%d error(s) in mie input-file, aborting\n", status);
    return status;
  }

  /* Do the Mie calculation */

  status = mie(Input,&Output);
  if ( status != 0.0 ) {
    fprintf (stderr, "Error during execution of mie\n");
    return status;
  }
 
  /* Print output ***FIXME*** THIS SHOULD BE USER DEFINED SOMEHOW */

  return status;

}

float yytext2float(char* yytext, int which_token, int max_token) {
  double dnumber;
  char token[] = " ";
  char *p, *s=NULL;
  int count;

  p = strtok(yytext,token);
  count = 1;
  while ( (p = strtok(NULL,token)) != NULL) {
    count++;
    if ( count == which_token ) {
      s = p;
    }
  }
  if ( (count != which_token) && (count>max_token) ) {
    fprintf(stderr,"Wrong number of arguments (%d) on line %d\n", count, line_number+1);
  }
  dnumber = atof(s);
  /*   printf("dnumner %f\n",dnumber);*/
  return (float)dnumber;
}

float* yytext2floats(char* yytext, int nary) {
  int i;
  char token[] = " ";
  char *p;

  /* Fill up ary */
  float *ary    = (float *) calloc (nary, sizeof(float));

  p = strtok(yytext,token);
  i = 0;
  while ( (p = strtok(NULL,token)) != NULL) {
    ary[i++] = atof(p);
  }
  return ary;
}

int yytext2nfloats(char* yytext) {
  int nary;
  char token[] = " ";
  char *p;

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext,token);
  while ( (p = strtok(NULL,token)) != NULL) {
    nary++;
  }
  return nary;
}

int* yytext2integers(char* yytext, int nary) {
  int i;
  char token[] = " ";
  char *p;

  /* Fill up ary */
  int *ary    = (int *) calloc (nary, sizeof(int));

  p = strtok(yytext,token);
  i = 0;
  while ( (p = strtok(NULL,token)) != NULL) {
    ary[i++] = atoi(p);
  }
  return ary;
}

int yytext2nintegers(char* yytext) {
  int nary;
  char token[] = " ";
  char *p;

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext,token);
  while ( (p = strtok(NULL,token)) != NULL) {
    nary++;
  }
  return nary;
}

int yytext2int(char* yytext, int which_token, int max_token) {
  int inumber;
  char token[] = " ";
  char *p, *s=NULL;
  int count;

  p = strtok(yytext,token);
  count = 1;
  while ( (p = strtok(NULL,token)) != NULL) {
    count++;
    if ( count == which_token ) {
      s = p;
    }
  }
  if ( (count != which_token) && (count>max_token) ) {
    fprintf(stderr,"Wrong number of arguments on line %d\n", line_number+1);
  }
  inumber = atoi(s);
  /*  printf("inumner %d\n",inumber);*/
  return inumber;
}

char* yytext2string(char* yytext, int which_token, int max_token) {
  char token[] = " ";
  char *p, *s=NULL, *string=NULL;
  int count;

  p = strtok(yytext,token);
  count = 1;
  while ( (p = strtok(NULL,token)) != NULL) {
    count++;
    if ( count == which_token ) {
      s = p;
    }
  }
  if ( (count != which_token) && (count>max_token) ) {
    fprintf(stderr,"Wrong number of arguments (%d) on line %d\n", count, line_number+1);
  }
  string = (char *) calloc (strlen(s)+1, sizeof (char));
  strcpy (string, s);
  return string;
}





/**************************************************/
/* Check the Mie input structure for plausibility */
/**************************************************/

int mie_check (input_struct input)
{
  int status=0;
  status =0;

  return status;
}
  
